from __future__ import print_function


__title__ = 'chanjo'
__version__ = 'v0.0.3-beta'
__author__ = 'Robin Andeer'
__license__ = 'MIT'
__copyright__ = 'Copyright 2013 Robin Andeer'
__docformat__ = 'restructuredtext'